--[[TODO
* Figure out what button clicking should do?
* Add little chat buttons.
* Color code borders
* Add events into the various addons that show damage text like Parrot, WoW's builtin one, the one I use, etc.
]]--

local Hermes = LibStub("AceAddon-3.0"):GetAddon("Hermes")
local MODNAME = "HermesDisplay"
local HermesDisplay = Hermes:NewModule(MODNAME, "AceEvent-3.0", "AceTimer-3.0")
local L = LibStub('AceLocale-3.0'):GetLocale('Hermes')

--LibButtonFacade support
local LBF  = LibStub("LibButtonFacade",true)
local LBFHermes = {} 
LBFHermes.Name = "Hermes"

--ToolTip
local QTip	= LibStub("LibQTip-1.0")
local ToolTip = nil

--frames
local LayoutFrame = nil
local LayoutFrameResizer = nil
local RecycledFrame = nil
local CommStatusFrame = nil
local INDICATOR_FRAME_COUNT = 1

--other
local SCAN_FREQUENCY = 0.1
local REALLY_BIG_NUMBER = 999999
HermesDisplay.PlayerZone = nil
HermesDisplay.Senders = nil
local SenderStatusTimer = nil
local SENDER_STATUS_UPDATE_FREQUENCY = 1
local ICONSIZE = 32
local ICONSIZE_COMM = ICONSIZE / 2
local ICON_STATUS_NOTREADY = "Interface\\RAIDFRAME\\ReadyCheck-NotReady.blp"
local ICON_STATUS_READY = "Interface\\RAIDFRAME\\ReadyCheck-Ready.blp"
local ICON_STATUS_UNKNOWN = "Interface\\RAIDFRAME\\ReadyCheck-Waiting.blp"
local IMAGE_RESIZE = "Interface\\Addons\\Hermes\\Textures\\Resize.tga"

local ICON_VOICE_SPEAKER = "Interface\\COMMON\\VOICECHAT-SPEAKER.BLP"
local ICON_VOICE_SPEAKER_WAVES = "Interface\\COMMON\\VOICECHAT-ON.BLP"
local ICON_VOICE_SPEAKER_MUTED = "Interface\\COMMON\\VOICECHAT-MUTED.BLP"

--------------------------------------------------------------------
-- Public Module Methods
--------------------------------------------------------------------
function HermesDisplay:AddSpell(spellid, classEnum)
	local name, _, icon, _, _, _, _, _, _ = GetSpellInfo(spellid)
	
	local spellInfo = {
		spellid = spellid,
		spellname = name,
		icon = icon,
		senders = {}
	}
				
	local button = self:FetchButton(icon)
	button.spellInfo = spellInfo
	button:ClearAllPoints()
	button:SetParent(LayoutFrame)
	local index = self:FindButtonIndex(RecycledFrame.Buttons, button)
	tremove(RecycledFrame.Buttons, index)
	tinsert(LayoutFrame.Buttons, button)

	self:UpdateButtonCooldowns(button)
	self:UpdateButtonScripts(button)
	self:UpdateButtonVisuals(button)

	self:ArrangeLayout()
	
	return spell
end

function HermesDisplay:RemoveSpell(spellid)
	for i, button in ipairs(LayoutFrame.Buttons) do
		if(tostring(button.spellInfo.spellid) == tostring(spellid)) then
			self:RecycleButton(button)
			return
		end
	end
end

function HermesDisplay:AddSender(senderName, classEnum)
	local sender = {
		classEnum = classEnum,
		raidId = nil,
		zone = nil,
		online = true,
		connected = true,
	}
	self.Senders[senderName] = sender
	self:UpdateSenderStatus(senderName)
	return sender
end

function HermesDisplay:RemoveSender(senderName)
	--remove all spell data for this sender
	HermesDisplay:RemoveSpellsForSender(senderName)
	
	--remove sender from senders
	HermesDisplay.Senders[senderName] = nil
end

function HermesDisplay:GetSender(senderName)
	for name, sender in pairs(HermesDisplay.Senders) do
		if(name == senderName) then
			return sender
		end
	end
	
	return nil
end

function HermesDisplay:AddSenderSpellInfo(senderName, spellid, duration)
	if(not self:GetSender(senderName)) then
		error("unknown sender in AddSenderSpellInfo")
	end
	
	for _, button in ipairs(LayoutFrame.Buttons) do
		if(button.spellInfo.spellid == spellid) then
			local senderIndex = self:FindSenderIndexInButton(button, senderName)
			local sender = nil
			if(not senderIndex) then
				sender = {
					senderName = senderName
				}
				tinsert(button.spellInfo.senders, sender)
			else
				sender = button.spellInfo.senders[senderIndex]
			end
			
			sender.duration = duration
			sender.timestamp = GetTime()
			
			self:UpdateButtonCooldowns(button)
			self:UpdateButtonScripts(button)
			self:UpdateButtonVisuals(button, false)
			
			if(sender.duration) then
				button:SetScript("OnUpdate", OnButtonUpdate) -- start the OnUpdate script if not already started
			end

			return
		end
	end
	
	--if we made it here, we're not tracking the spell
end

function HermesDisplay:RemoveSpellsForSender(senderName)
	for _, button in ipairs(LayoutFrame.Buttons) do
		local senderIndex = self:FindSenderIndexInButton(button, senderName)
		if(senderIndex) then
			tremove(button.spellInfo.senders, senderIndex)
			
			HermesDisplay:UpdateButtonCooldowns(button)
			HermesDisplay:UpdateButtonScripts(button)
			HermesDisplay:UpdateButtonVisuals(button)
		end
	end
end

--------------------------------------------------------------------
-- Sender Helpers
--------------------------------------------------------------------
function HermesDisplay:UpdateSenderStatus(senderName)
	local sender = self.Senders[senderName]
	if(Hermes.PlayerInRaid) then
		--if we're in a raid
		sender.raidId = UnitInRaid(senderName)
		if(sender.raidId) then
			local oldZone = sender.zone
			local oldOnline = sender.online
			local oldDead = sender.dead
			
			local name, _, _, _, _, _, zone, online, dead, _, _ = GetRaidRosterInfo(sender.raidId)
			
			if(name == senderName) then
				sender.zone = zone
				sender.online = online
				sender.dead = dead
			else
				error("Unexpected sender status during UpdateSenderStatus")
			end
			
			if(oldZone ~= sender.zone or oldOnline ~= sender.online or oldDead ~= sender.dead) then
				HermesDisplay:UpdateButtonVisuals(nil)
			end
		else
			--sender dropped from raid
			self:RemoveSender(senderName)
		end
	elseif(Hermes.PlayerInParty) then
		--if we're in a party
		if(UnitInParty(senderName)) then
			local oldDead = sender.dead
			
			sender.zone = HermesDisplay.PlayerZone --there is no API to get the zone of a player or party member so assume zones match
			sender.online = 1 --there is no API to get the online status of a player or party member so assume online
			
			local dead = UnitIsDead(senderName) or UnitIsGhost(senderName)
			if(dead == true) then
				sender.dead = 1
			else
				sender.dead = nil
			end
			
			if(oldDead ~= sender.dead) then
				HermesDisplay:UpdateButtonVisuals(nil)
			end
		else
			--sender dropped from party
			self:RemoveSender(senderName)
		end
	end
end

function HermesDisplay:OnUpdateSenderStatus()
	if(HermesDisplay.Senders) then
		for senderName, senderInfo in pairs(HermesDisplay.Senders) do
			self:UpdateSenderStatus(senderName)
		end
	end
end

function HermesDisplay:KillSenderStatusTimer()
	if(SenderStatusTimer) then
		self:CancelTimer(SenderStatusTimer, true)
		SenderStatusTimer = self:ScheduleTimer("OnUpdateSenderStatus", 0.1, nil) --convert to one shot timer
		self:CancelTimer(SenderStatusTimer, true) --cancel again, we don't care if it ran or not
		SenderStatusTimer = nil
	end
end

function HermesDisplay:StartSenderStatusTimer()
	SenderStatusTimer = self:ScheduleRepeatingTimer("OnUpdateSenderStatus", SENDER_STATUS_UPDATE_FREQUENCY, nil)
end

--------------------------------------------------------------------
-- LibButtonFacade
--------------------------------------------------------------------
function LBFHermes.SetupGroup()
	local group = LBF:Group(LBFHermes.Name, MODNAME);
	if(not Hermes.db.profile.display.lbf) then
		Hermes.db.profile.display.lbf = { skin = "Blizzard", gloss = 0, backdrop = false, colors = {} }
	end
	
	group:Skin(
		Hermes.db.profile.display.lbf.skin,
		Hermes.db.profile.display.lbf.gloss,
		Hermes.db.profile.display.lbf.backdrop,
		Hermes.db.profile.display.lbf.colors)
end

function LBFHermes.RegisterSkin(arg, skin, gloss, backdrop, group, button, colors)
	if(group and group == MODNAME) then
		Hermes.db.profile.display.lbf.skin = skin
		Hermes.db.profile.display.lbf.gloss = gloss
		Hermes.db.profile.display.lbf.backdrop = backdrop
		Hermes.db.profile.display.lbf.colors = colors
	end
end

--------------------------------------------------------------------
-- AceAddon
--------------------------------------------------------------------
function HermesDisplay:OnEnable()
	
end

function HermesDisplay:OnDisable()
	
end

function HermesDisplay:SetupLBFGroup()
	LBFHermes.SetupGroup()
end

function HermesDisplay:OnInitialize()
	self:CreateRecycledFrame()
	self:CreateLayoutFrame()
	--self:CreateCommStatusFrame()
	self:SetEnabledState(true)
	HermesDisplay.Senders = {}
	
	if(LBF) then
		self:SetupLBFGroup()
		LBF:RegisterSkinCallback(LBFHermes.Name, LBFHermes.RegisterSkin, LBFHermes)
	end
end

function HermesDisplay:EnableEvents(toggle)						--enables message and event handling
	if(toggle == true) then
		self:RegisterEvent("PLAYER_ENTERING_WORLD")
		self:RegisterEvent("ZONE_CHANGED")
		self:RegisterEvent("ZONE_CHANGED_NEW_AREA")
		self:RegisterEvent("ZONE_CHANGED_INDOORS")
		self:StartSenderStatusTimer()
	else
		self:UnregisterEvent("PLAYER_ENTERING_WORLD")
		self:UnregisterEvent("ZONE_CHANGED")
		self:UnregisterEvent("ZONE_CHANGED_NEW_AREA")
		self:UnregisterEvent("ZONE_CHANGED_INDOORS")
		self:KillSenderStatusTimer()
	end
end

function HermesDisplay:Show(toggle)								--toggles showing of the window
	if(toggle == true) then
		self:RestoreLayoutFramePosition()
		LayoutFrame:Show()
	else
		LayoutFrame:Hide()
	end
end

function HermesDisplay:ResetSenders()							--wipe out all senders and button state of senders, but keeps buttons
	if(HermesDisplay.Senders) then
		for senderName, senderInfo in pairs(HermesDisplay.Senders) do
			HermesDisplay:RemoveSender(senderName)
		end

		wipe(HermesDisplay.Senders)
		HermesDisplay.Senders = nil
	end
end

function HermesDisplay:ResetAll()								--resets everything to a clean slate (for example as needed due to a profile change)
	self:ResetSenders()
	self:RecycleButtons()
end

function HermesDisplay:Shutdown()
	HermesDisplay:EnableEvents(false)
	HermesDisplay:ResetSenders()
end

function HermesDisplay:Startup()								--initializes data
	self:Show(false) --hide by default
	self:RestoreLayoutFramePosition()
	if(not HermesDisplay.Senders) then
		HermesDisplay.Senders = {}
	end
	self.PlayerZone = GetRealZoneText()
	HermesDisplay:EnableEvents(true)
end

function HermesDisplay:PLAYER_ENTERING_WORLD()				--used only for one time player initialization that has to happen after player is logged in.
	self:UnregisterEvent("PLAYER_ENTERING_WORLD")
	self.PlayerZone = GetRealZoneText()
end

function HermesDisplay:ZONE_CHANGED()
	self.PlayerZone = GetRealZoneText()
end

function HermesDisplay:ZONE_CHANGED_NEW_AREA()
	self.PlayerZone = GetRealZoneText()
end

function HermesDisplay:ZONE_CHANGED_INDOORS()
	self.PlayerZone = GetRealZoneText()
end

--------------------------------------------------------------------
-- Recycled Frame
--------------------------------------------------------------------
function HermesDisplay:CreateRecycledFrame()
	RecycledFrame = CreateFrame("Frame", "HermesDisplayRecycledFrame", UIParent)
	RecycledFrame:SetPoint("CENTER")
	RecycledFrame:SetWidth(50)
	RecycledFrame:SetHeight(50)
	RecycledFrame:Hide()

	RecycledFrame:EnableMouse(false)
	RecycledFrame:SetMovable(false)
	RecycledFrame:SetToplevel(false)
	RecycledFrame.Buttons = {}
end

--------------------------------------------------------------------
-- LayoutFrame
--------------------------------------------------------------------
local function OnLayoutFrameDragStart(self, button)
	if(Hermes.db.profile.display.locked == false) then
		self:StartMoving();
	end
end

local function OnLayoutFrameDragStop(self, button)
	self:StopMovingOrSizing();
	HermesDisplay:SaveLayoutFramePosition()
end

local function OnLayoutFrameSizeChanged(self, width, height)
	HermesDisplay:ArrangeLayout()
	HermesDisplay:SaveLayoutFramePosition()
end

function HermesDisplay:CreateLayoutFrame()
	local scale = Hermes.db.profile.display.scale
	local locked = Hermes.db.profile.display.locked
	local padding = Hermes.db.profile.display.padding

	--Icon Frame
	LayoutFrame = CreateFrame("Frame", "HermesDisplayLayoutFrame", UIParent)
	LayoutFrame:SetPoint("CENTER", UIParent, "CENTER")
	LayoutFrame:SetScale(scale)
	LayoutFrame:EnableMouse(true)
	LayoutFrame:SetMovable(true)
	LayoutFrame:SetResizable(true)
	LayoutFrame:SetUserPlaced(false)
	LayoutFrame:SetMinResize(ICONSIZE + padding + 5, ICONSIZE + padding + 5)
	LayoutFrame:SetToplevel(false)
	LayoutFrame:SetWidth(250)
	LayoutFrame:SetHeight(100)
	LayoutFrame:RegisterForDrag("LeftButton");
	LayoutFrame:SetScript("OnDragStart", OnLayoutFrameDragStart)
	LayoutFrame:SetScript("OnDragStop", OnLayoutFrameDragStop)
	LayoutFrame:SetScript("OnSizeChanged", OnLayoutFrameSizeChanged)
	LayoutFrame:SetBackdrop({
			bgFile = "Interface\\Tooltips\\UI-Tooltip-Background", 
			tile = false,
			tileSize = 32,
			edgeSize = 10,
			insets = {
				left = -1,
				right = -1,
				top = -1,
				bottom = -1
			}
		})
		
	if(locked == false) then
		LayoutFrame:SetBackdropColor(
			0,
			0,
			0,
			0.5)
	else
		LayoutFrame:SetBackdropColor(
			0,
			0,
			0,
			0)
	end
	LayoutFrame.Buttons = {}
	LayoutFrame:Hide()
	
	LayoutFrameResizer = CreateFrame("Frame", "HermesReceiverLayoutFrameResizer", LayoutFrame)
	LayoutFrameResizer:SetWidth(20)
	LayoutFrameResizer:SetHeight(20)
	LayoutFrameResizer:EnableMouse(true)
	LayoutFrameResizer:RegisterForDrag("LeftButton")
	LayoutFrameResizer:SetPoint("BOTTOMRIGHT")
	LayoutFrameResizer:SetScript("OnMouseDown", function() LayoutFrame:StartSizing("BOTTOMRIGHT") end)
	LayoutFrameResizer:SetScript("OnMouseUp", function() LayoutFrame:StopMovingOrSizing() end)
	--resize gripper texture
	LayoutFrameResizer.ResizeTexture = LayoutFrameResizer:CreateTexture()
	LayoutFrameResizer.ResizeTexture:SetTexture(IMAGE_RESIZE)
	LayoutFrameResizer.ResizeTexture:SetDrawLayer("OVERLAY")
	LayoutFrameResizer.ResizeTexture:SetAllPoints()
	HermesDisplay:UpdateFrameResizerPosition()
	
	if(Hermes.db.profile.display.locked == false) then
		LayoutFrameResizer:Show()
	else
		LayoutFrameResizer:Hide()
	end
end

function HermesDisplay:Lock(lock)
	if(lock == false) then
		LayoutFrame:SetBackdropColor(
			0,
			0,
			0,
			0.5)
		LayoutFrameResizer:Show()
	else
		LayoutFrame:SetBackdropColor(
			0,
			0,
			0,
			0)
		LayoutFrameResizer:Hide()
	end
end

function HermesDisplay:SetWindowScale(scale)
	LayoutFrame:SetScale(scale)
end

function HermesDisplay:UpdateFrameResizerPosition()
	local anchor
	local buttonanchor = Hermes.db.profile.display.buttonanchor
	if buttonanchor == "TOPLEFT" then --good
		anchor = "BOTTOMRIGHT"
		RotateTexture(LayoutFrameResizer.ResizeTexture, 0)
	elseif buttonanchor == "TOPRIGHT" then --good
		anchor = "BOTTOMLEFT"
		RotateTexture(LayoutFrameResizer.ResizeTexture, -90)
	elseif buttonanchor == "BOTTOMLEFT" then --good
		anchor = "TOPRIGHT"
		RotateTexture(LayoutFrameResizer.ResizeTexture, 90)
	elseif buttonanchor == "BOTTOMRIGHT" then
		anchor = "TOPLEFT"
		RotateTexture(LayoutFrameResizer.ResizeTexture, 180)
	else
		anchor = "BOTTOMRIGHT"
		RotateTexture(LayoutFrameResizer.ResizeTexture, 0)
	end

	
	LayoutFrameResizer:ClearAllPoints()
	LayoutFrameResizer:SetPoint(anchor)
	LayoutFrameResizer:SetScript("OnMouseDown", function() LayoutFrame:StartSizing(anchor) end)
end

function HermesDisplay:ArrangeLayout()
	local padding = Hermes.db.profile.display.padding
	local buttonanchor = Hermes.db.profile.display.buttonanchor
	
	local anchor, multy, multx

	if buttonanchor == "TOPLEFT" then
		anchor = "TOPLEFT"
		multy = -1
		multx = 1
	elseif buttonanchor == "TOPRIGHT" then
		anchor = "TOPRIGHT"
		multy = -1
		multx = -1
	elseif buttonanchor == "BOTTOMLEFT" then
		anchor = "BOTTOMLEFT"
		multy = 1
		multx = 1
	elseif buttonanchor == "BOTTOMRIGHT" then
		anchor = "BOTTOMRIGHT"
		multy = 1
		multx = -1
	else
		anchor = "TOPLEFT"
		multy = -1
		multx = 1
	end
	
	local columnsPerRow = floor(LayoutFrame:GetWidth() / (ICONSIZE + (padding)))
	if(columnsPerRow <= 0) then
		columnsPerRow = 1
	end
	
	--sort the spells by spell name
	table.sort(LayoutFrame.Buttons,
		function(a, b)
			return a.spellInfo.spellname < b.spellInfo.spellname
		end
	)
	
	local row = 0
	local column = 0
	for i, button in ipairs(LayoutFrame.Buttons) do
		local xoffset = ((column) * (ICONSIZE + padding)) * multx
		local yoffset = ((row) * (ICONSIZE + padding)) * multy
		
		button:Hide()
		button:ClearAllPoints()
		button:SetPoint(anchor, LayoutFrame, anchor, xoffset, yoffset)
		button:SetHeight(ICONSIZE)
		button:SetWidth(ICONSIZE)
		button:Show()

		column = column + 1
		if(column == columnsPerRow) then
			row = row + 1
			column = 0
		end
	end
	
	-- LBF ANONYMOUS SKIN SUPPORT
	if(LBF) then
		LBF:Group(LBFHermes.Name, MODNAME):ReSkin() --resizing buttons requires reskinning...
	end
end

function HermesDisplay:SaveLayoutFramePosition()
	--only update bar frame X, Y if it's unanchored. This way we don't reset it's position to the location of the window it's anchored to
	Hermes.db.profile.display.x = LayoutFrame:GetLeft()
	Hermes.db.profile.display.y = LayoutFrame:GetTop()
	Hermes.db.profile.display.w = LayoutFrame:GetWidth()
	Hermes.db.profile.display.h = LayoutFrame:GetHeight()
	LayoutFrame:SetUserPlaced(false)
end

function HermesDisplay:RestoreLayoutFramePosition()				--sets the starting point for all window positions based on profile settings (if it exists)
	if(not Hermes.db.profile.display.x or not Hermes.db.profile.display.y) then
		LayoutFrame:ClearAllPoints()
		LayoutFrame:SetPoint("CENTER", UIParent, "CENTER")
		Hermes.db.profile.display.x = LayoutFrame:GetLeft()
		Hermes.db.profile.display.y = LayoutFrame:GetTop()
		Hermes.db.profile.display.w = 250
		Hermes.db.profile.display.h = 100
	end
	
	LayoutFrame:ClearAllPoints()
	LayoutFrame:SetPoint("TOPLEFT", UIParent, "BOTTOMLEFT",
		Hermes.db.profile.display.x,
		Hermes.db.profile.display.y
	)
	LayoutFrame:SetWidth(Hermes.db.profile.display.w)
	LayoutFrame:SetHeight(Hermes.db.profile.display.h)
	LayoutFrame:SetUserPlaced(nil)
end

--------------------------------------------------------------------
-- CommStatus Frame
--------------------------------------------------------------------
--[[
function HermesDisplay:CreateCommStatusFrame()
	local scale = Hermes.db.profile.display.scale

	CommStatusFrame = CreateFrame("Frame", "HermesCommStatusFrame", LayoutFrame)
	--CommStatusFrame:SetAllPoints()
	--CommStatusFrame:SetScale(scale)
	CommStatusFrame:EnableMouse(true)
	CommStatusFrame:SetMovable(true)
	CommStatusFrame:SetResizable(false)
	CommStatusFrame:SetUserPlaced(false)
	CommStatusFrame:SetWidth(ICONSIZE_COMM * 2)
	CommStatusFrame:SetHeight(ICONSIZE_COMM)
	
	CommStatusFrame.Send = CreateFrame("Frame", "HermesCommStatusFrameSend", CommStatusFrame)
	CommStatusFrame.Send:SetPoint("LEFT", CommStatusFrame, "LEFT")
	CommStatusFrame.Send:SetPoint("TOP", CommStatusFrame, "TOP")
	CommStatusFrame.Send:SetPoint("BOTTOM", CommStatusFrame, "BOTTOM")
	CommStatusFrame.Send:SetWidth(ICONSIZE_COMM)
	
	CommStatusFrame.Send.Backgound = CommStatusFrame.Send:CreateTexture(nil, "BACKGROUND")
	CommStatusFrame.Send.Artwork = CommStatusFrame.Send:CreateTexture(nil, "ARTWORK")
	CommStatusFrame.Send.Overlay = CommStatusFrame.Send:CreateTexture(nil, "OVERLAY")
	
	CommStatusFrame.Send.Backgound:SetTexture(ICON_VOICE_SPEAKER)
	CommStatusFrame.Send.Artwork:SetTexture(ICON_VOICE_SPEAKER_WAVES)
	CommStatusFrame.Send.Overlay:SetTexture(ICON_VOICE_SPEAKER_MUTED)
	
	CommStatusFrame.Send.Backgound:SetAllPoints()
	CommStatusFrame.Send.Artwork:SetAllPoints()
	CommStatusFrame.Send.Overlay:SetAllPoints()
	
	self:AnchorCommStatusFrame()
end
]]--
--[[
function HermesDisplay:AnchorCommStatusFrame()
	local buttonanchor = Hermes.db.profile.display.buttonanchor
	CommStatusFrame:ClearAllPoints()
	
	if buttonanchor == "TOPLEFT" then
		CommStatusFrame:SetPoint("BOTTOMLEFT", LayoutFrame, "TOPLEFT", 0, 0)
	elseif buttonanchor == "TOPRIGHT" then
		CommStatusFrame:SetPoint("BOTTOMRIGHT", LayoutFrame, "TOPRIGHT", 0, 0)
	elseif buttonanchor == "BOTTOMLEFT" then
		CommStatusFrame:SetPoint("TOPLEFT", LayoutFrame, "BOTTOMLEFT", 0, 0)
	elseif buttonanchor == "BOTTOMRIGHT" then
		CommStatusFrame:SetPoint("TOPRIGHT", LayoutFrame, "BOTTOMRIGHT", 0, 0)
	else
		CommStatusFrame:SetPoint("BOTTOMLEFT", LayoutFrame, "TOPLEFT", 0, 0)
	end
end
]]--
--------------------------------------------------------------------
-- Buttons
--------------------------------------------------------------------
function HermesDisplay:CreateButton()
	--button
	local button = CreateFrame("Button", "HermesDisplayButton"..tostring(INDICATOR_FRAME_COUNT), RecycledFrame, "ActionButtonTemplate")
	button.id = INDICATOR_FRAME_COUNT
	
	button.icon = _G[button:GetName().."Icon"]
	button.flash = _G[button:GetName().."Flash"]
	button.count = _G[button:GetName().."Count"]
	button.name = _G[button:GetName().."Name"]
	button.border = _G[button:GetName().."Border"]
	button.cooldown = _G[button:GetName().."Cooldown"]
	button.normaltexture = _G[button:GetName().."NormalTexture"]
	button.pushedtexture = _G[button:GetName().."PushedTexture"]
	button.highlighttexture = _G[button:GetName().."HighlightTexture"]
	button.checkedtexture = _G[button:GetName().."CheckedTexture"]
	
	button:SetAllPoints()
	button:EnableMouse(true)
	button:SetMovable(true)
	button:SetResizable(true)
	button:RegisterForClicks("AnyUp")
	button:SetMotionScriptsWhileDisabled(true)
	
	tinsert(RecycledFrame.Buttons, button)
	INDICATOR_FRAME_COUNT = INDICATOR_FRAME_COUNT + 1

	--LBF SUPPORT
	if(LBF) then
		LBF:Group(LBFHermes.Name, MODNAME):AddButton(button)
	end
	
	return button
end

function HermesDisplay:ResetButton(button)
	button.spellInfo = nil
	button.LastScan = 0

	button:SetScript("OnMouseUp", nil)
	button:SetScript("OnEnter", nil)
	button:SetScript("OnLeave", nil)
	button:SetScript("OnUpdate", nil)
	
	button:SetButtonState("DISABLED", true)
	button.icon:SetTexture("")
	button.icon:SetDesaturated(true)
	button.icon:SetVertexColor(
		Hermes.db.profile.display.VertexColorDoesntExist[1],
		Hermes.db.profile.display.VertexColorDoesntExist[2],
		Hermes.db.profile.display.VertexColorDoesntExist[3],
		Hermes.db.profile.display.VertexColorDoesntExist[4])
end

function HermesDisplay:FetchButton(texture)
	local button
	if(RecycledFrame.Buttons and #RecycledFrame.Buttons > 0) then
		button = RecycledFrame.Buttons[1]
	else
		button = self:CreateButton()
	end
	
	HermesDisplay:ResetButton(button)
	
	button:SetScript("OnMouseUp", OnButtonMouseUp)
	button:SetScript("OnEnter", OnButtonEnter)
	button:SetScript("OnLeave", OnButtonLeave)
	button:SetScript("OnUpdate", OnButtonUpdate)

	button.icon:SetTexture(texture)
	button.icon:SetDesaturated(true)
	button.icon:SetVertexColor(0.3, 0.3, 0.3, 1)
	
	return button
end

function HermesDisplay:FindButtonIndex(buttons, button)
	for index, i in ipairs(buttons) do
		if(i.id == button.id) then
			return index
		end
	end

	return nil
end

function GetTableCount(tbl)
	local cnt = 0
	for k, v in pairs(tbl) do
		cnt = cnt + 1
	end
	return cnt
end

function HermesDisplay:CalculateButtonSenderStateInfo(button)
	local totalsenders = 0
	local availableSenders = 0
	local mindurationAvailableSenders = nil
	local unavailableSenders = 0
	local mindurationUnavailableSenders = nil
	
	for i, senderInfo in ipairs(button.spellInfo.senders) do
		totalsenders = totalsenders + 1
		local sender = self:GetSender(senderInfo.senderName);
		if(sender.online == 1 and not sender.dead and sender.zone == self.PlayerZone) then --sender is available
			if(senderInfo.duration == nil) then
				--increment available count
				availableSenders = availableSenders + 1
			else
				if(not mindurationAvailableSenders or senderInfo.duration < mindurationAvailableSenders) then
					--set minimum duration
					mindurationAvailableSenders = senderInfo.duration
				end
			end
		elseif(sender.online == 0 or sender.dead or sender.zone ~= self.PlayerZone) then --sender not available
			if(senderInfo.duration == nil) then
				--increment unavailable count
				unavailableSenders = unavailableSenders + 1
			else
				if(not mindurationUnavailableSenders or senderInfo.duration < mindurationUnavailableSenders) then
					--set minimum duration
					mindurationUnavailableSenders = senderInfo.duration
				end
			end
		end
	end
	
	return totalsenders, availableSenders, mindurationAvailableSenders, unavailableSenders, mindurationUnavailableSenders
end

function HermesDisplay:SetButtonStyleAvailable(button, availableCount, minduration, showOverlay, killtimer)
	button:SetButtonState("NORMAL", false)
	button.icon:SetDesaturated(false)
	
	if(showOverlay) then
		button.icon:SetVertexColor(
			Hermes.db.profile.display.VertexColorUnavailable[1],
			Hermes.db.profile.display.VertexColorUnavailable[2],
			Hermes.db.profile.display.VertexColorUnavailable[3],
			Hermes.db.profile.display.VertexColorUnavailable[4])
	else
		button.icon:SetVertexColor(1, 1, 1, 1)
	end
	
	--this is a straight up hack. I want to make sure that any final "sparks" are shown when a cooldown finishes
	--but I also want to make sure that we stop timers appropriately too. I'm going on the assumption that sparks will
	--never last more than 1 second in duration, and when you factor in lag, update frequency, etc, I'm doubling it.
	--Therefore, if the minDuration is > 2 seconds, don't kill it, otherwise do...
	if(minduration and minduration > 2 or killtimer and killtimer == true) then
		CooldownFrame_SetTimer(button.cooldown, 0, 0, 0) --stop cooldown if it's running
	end
	
	--now update the text
	button.count:SetText(tostring(availableCount))
end

function HermesDisplay:SetButtonStyleUnavailable(button)
	button:SetButtonState("DISABLED", true)
	button.icon:SetDesaturated(true)
	button.icon:SetVertexColor(
		Hermes.db.profile.display.VertexColorDoesntExist[1],
		Hermes.db.profile.display.VertexColorDoesntExist[2],
		Hermes.db.profile.display.VertexColorDoesntExist[3],
		Hermes.db.profile.display.VertexColorDoesntExist[4])
	button.count:SetText(nil)
	CooldownFrame_SetTimer(button.cooldown, 0, 0, 0) --stop cooldown if it's running
end

function HermesDisplay:SetButtonStyleOnCooldown(button, duration, showOverlay)
	button:SetButtonState("NORMAL", false)
	button.icon:SetDesaturated(false)
	
	if(showOverlay) then
		button.icon:SetVertexColor(
			Hermes.db.profile.display.VertexColorUnavailable[1],
			Hermes.db.profile.display.VertexColorUnavailable[2],
			Hermes.db.profile.display.VertexColorUnavailable[3],
			Hermes.db.profile.display.VertexColorUnavailable[4])
	else
		button.icon:SetVertexColor(1, 1, 1, 1)
	end
	
	button.count:SetText(nil)
	
	--ensure duration is actually a value
	if(duration) then
		CooldownFrame_SetTimer(button.cooldown, GetTime(), duration, 1) --start/reset cooldown
	end
end

function HermesDisplay:UpdateButtonVisuals(button)
	if(button) then
		local totalsenders, availableSenders, mindurationAvailableSenders, unavailableSenders, mindurationUnavailableSenders = HermesDisplay:CalculateButtonSenderStateInfo(button)
		if(totalsenders > 0) then
			if(availableSenders == 0) then
				--we may or may not be on cooldown, it might be that we have some unavailable senders instead
				if(mindurationAvailableSenders ~= nil) then
					--there is an active sender with a cooldown, so go ahead and start a cooldown
					HermesDisplay:SetButtonStyleOnCooldown(button, mindurationAvailableSenders, false)
				else
					--There aren't any active senders with a cooldown running, yet totalsenders is > 0.
					--This can only mean that there are inactive senders. We don't care about their status though.
					--This will reset button count to 0, create a "dark" overlay, and will reset the timer
					HermesDisplay:SetButtonStyleAvailable(button, "", nil, true, true)
				end
			else
				HermesDisplay:SetButtonStyleAvailable(button, availableSenders, mindurationAvailableSenders, false)
			end
		else
			HermesDisplay:SetButtonStyleUnavailable(button)
		end
	else
		for _, i in ipairs(LayoutFrame.Buttons) do
			self:UpdateButtonVisuals(i)
		end
	end
end

function HermesDisplay:UpdateButtonScripts(button)		--resets the OnUpdate script is necessary when no cooldowns are running
	for i, senderInfo in ipairs(button.spellInfo.senders) do
		if(senderInfo.duration) then
			return
		end
	end
	button:SetScript("OnUpdate", nil)
end

function HermesDisplay:UpdateButtonCooldowns(button)	--this method is called to decrement the duration of each sender since it was last called
	for i, senderInfo in ipairs(button.spellInfo.senders) do
		--if the sender has a cooldown value (counting down)
		if(senderInfo.duration) then
			--update the cooldown
			local elapsed = GetTime() - senderInfo.timestamp
			senderInfo.duration = senderInfo.duration - elapsed
			senderInfo.timestamp = GetTime() --set the timestamp too
			--if cooldown expired, then update the button
			if(senderInfo.duration <= 0) then
				senderInfo.duration = nil --reset sender cooldown
				self:UpdateButtonVisuals(button) --update button to reflect that cooldown is finished
			end
		end
	end
end

function OnButtonUpdate(self, elapsed)
	self.LastScan = self.LastScan + elapsed;
	while (self.LastScan > SCAN_FREQUENCY) do
		self.LastScan = self.LastScan - SCAN_FREQUENCY
		HermesDisplay:UpdateButtonCooldowns(self)
		HermesDisplay:UpdateButtonScripts(self)
		if(ToolTip and ToolTip.button and ToolTip.button == self) then
			--refresh the tooltip since it's being displayed for this button
			HermesDisplay:RefreshTooltip()
		end
	end
end

function OnButtonEnter(self, button)
	HermesDisplay:ShowTooltip(self, self)
end

function OnButtonLeave(self, button)
	QTip:Release(ToolTip)
	ToolTip = nil
end

function OnButtonMouseUp(self, button)
	wipe(self.spellInfo.senders)
	HermesDisplay:UpdateButtonCooldowns(self)
	HermesDisplay:UpdateButtonScripts(self)
	HermesDisplay:UpdateButtonVisuals(self)
	if(Hermes:IsReceiving()) then --prevent trying to send messages when not in raid or party. Otherwise Lua error
		Hermes:RequestSpellUpdate(self.spellInfo.spellid)
	end
end

function HermesDisplay:FindSenderIndexInButton(button, senderName)
	for i, sender in ipairs(button.spellInfo.senders) do
		if(sender.senderName == senderName) then
			return i
		end
	end
	
	return nil
end

function HermesDisplay:RecycleButton(button)
	local index = self:FindButtonIndex(LayoutFrame.Buttons, button)
	tremove(LayoutFrame.Buttons, index)
	tinsert(RecycledFrame.Buttons, button)
	self:ResetButton(button)
	button:SetParent(RecycledFrame)
	button:SetAllPoints()
	self:ArrangeLayout()
end

function HermesDisplay:RecycleButtons()
	if LayoutFrame and LayoutFrame.Buttons then
		while #LayoutFrame.Buttons > 0 do
			HermesDisplay:RecycleButton(LayoutFrame.Buttons[1])
		end
	end
end

--------------------------------------------------------------------
-- ToolTip
--------------------------------------------------------------------
local function SortByName(a, b)
	return a.senderName < b.senderName --sort by name if duration is the same
end

local function SortByDuration(a, b)
	local durationA 
	local durationB
	
	if(not a.duration) then durationA = 0 else durationA = a.duration end
	if(not b.duration) then durationB = 0 else durationB = b.duration  end
	
	if(durationA == durationB) then
		return SortByName(a, b)
	else
		return durationA < durationB --sort by duration
	end
end

local function SortByAlive(a, b)
	local senderA = HermesDisplay:GetSender(a.senderName)
	local senderB = HermesDisplay:GetSender(b.senderName)
	
	--alive
	if(not senderA.alive and senderB.alive) then
		return false --A below B
	end
	if(senderA.alive and not senderB.alive) then
		return true --A above B
	end
	if(not senderA.alive and not senderB.alive) then --they're both dead, sort by duration
		return SortByDuration(a, b)
	end
	
	--they're both alive, sort by duration
	return SortByDuration(a, b)
end

local function SortByZone(a, b)
	local senderA = HermesDisplay:GetSender(a.senderName)
	local senderB = HermesDisplay:GetSender(b.senderName)
	
	--zone
	if(senderA.zone ~= HermesDisplay.PlayerZone and senderB.zone == HermesDisplay.PlayerZone) then
		return false --A below B
	end
	if(senderA.zone == HermesDisplay.PlayerZone and senderB.zone ~= HermesDisplay.PlayerZone) then
		return true --A above B
	end
	if(senderA.zone ~= HermesDisplay.PlayerZone and senderB.zone ~= HermesDisplay.PlayerZone) then
		return SortByDuration(a, b)
	end
	
	--they're both in same zone as player, sort by duration
	return SortByAlive(a, b)
end

local function SortByConnected(a, b)
	local senderA = HermesDisplay:GetSender(a.senderName)
	local senderB = HermesDisplay:GetSender(b.senderName)
	
	--online
	if(not senderA.online and senderB.online) then
		return false --A below B
	end
	if(senderA.online and not senderB.online) then
		return true --A above B
	end
	if(not senderA.online and not senderB.online) then
		return SortByName(a, b)
	end

	--they're both connected, sort by duration
	return SortByZone(a, b)
end

local function SortSendersForToolTip(a, b)
	--returning false means A below B
	--returning true means A above B
	return SortByConnected(a, b)
end

function HermesDisplay:RefreshTooltip()
	if(ToolTip ~= nil) then
		ToolTip:Clear()
		ToolTip:AddHeader(ToolTip.button.spellInfo.spellname, " ", " ", " ", " ")
		
		if(#ToolTip.button.spellInfo.senders == 0) then
			local line = ToolTip:AddLine(
				L["None found"],
				nil,
				nil,
				nil,
				nil
			)
		else
			ToolTip:AddLine(
				L["Player"],
				L["Time"],
				L["Alive"],
				L["Conn"],
				L["Zone"])
			ToolTip:AddSeparator(1)
		
			table.sort(ToolTip.button.spellInfo.senders, function(a, b) return SortSendersForToolTip(a, b) end)
			
			for i, senderInfo in ipairs(ToolTip.button.spellInfo.senders) do
				local sender = HermesDisplay:GetSender(senderInfo.senderName)
				
				--available, playername, duration, alive, zone, online
				local iconAvailable
				if(not senderInfo.duration and not sender.dead and sender.connected and sender.zone == self.PlayerZone) then
					iconAvailable = "|T"..ICON_STATUS_READY..":0:0:0:0|t"
				else
					iconAvailable = "|T"..ICON_STATUS_NOTREADY..":0:0:0:0|t"
				end
				
				local iconDead
				if(sender.dead) then
					iconDead = "|T"..ICON_STATUS_NOTREADY..":0:0:0:0|t"
				else
					iconDead = "|T"..ICON_STATUS_READY..":0:0:0:0|t"
				end
				
				local iconOnline
				if(sender.online) then
					iconOnline = "|T"..ICON_STATUS_READY..":0:0:0:0|t"
				else
					iconOnline = "|T"..ICON_STATUS_NOTREADY..":0:0:0:0|t"
				end
				
				local iconZone
				if(sender.zone == self.PlayerZone) then
					iconZone = "|T"..ICON_STATUS_READY..":0:0:0:0|t"
				else
					iconZone = "|T"..ICON_STATUS_NOTREADY..":0:0:0:0|t"
				end
				
				local durationText
				if(sender.dead or not sender.connected or sender.zone ~= self.PlayerZone) then
					--special case if the user is not connected, dead, or in wrong zone
					if(senderInfo.duration) then
						--durationText = "|cFFFF0000"..tostring(floor(senderInfo.duration)).."|r" --red
						durationText = "|cFFFF0000"..self:SecondsToClock(senderInfo.duration).."|r" --red
					else
						durationText = "|cFFFF0000".."0".."|r" --red
					end
				elseif(senderInfo.duration == nil) then
					durationText = "|cFF00FF00".."0".."|r"									--green
				elseif(senderInfo.duration < 10) then
					--durationText = "|cFFFFFF00"..tostring(floor(senderInfo.duration)).."|r" --yellow
					durationText = "|cFFFFFF00"..self:SecondsToClock(senderInfo.duration).."|r" --yellow
				else
					--durationText = "|cFFFF0000"..tostring(floor(senderInfo.duration)).."|r" --red
					durationText = "|cFFFF0000"..self:SecondsToClock(senderInfo.duration).."|r" --red
				end
				
				local line = ToolTip:AddLine(
					--iconAvailable,
					Hermes:GetColorCodedClassText(senderInfo.senderName, sender.classEnum),
					durationText,
					iconDead,
					iconOnline,
					iconZone
				)
			end
		end
	end
end

function HermesDisplay:ShowTooltip(button, anchor)
	--if(ToolTip == nil and #button.spellInfo.senders > 0) then
	if(ToolTip == nil) then
		ToolTip = QTip:Acquire(MODNAME.." Tooltip", 5, "LEFT", "RIGHT", "RIGHT", "RIGHT", "RIGHT", "RIGHT")
		ToolTip:ClearAllPoints()
		ToolTip:SmartAnchorTo(anchor)
		ToolTip.button = button --used by each button during on update to refresh the tip
		self:RefreshTooltip()
		ToolTip:Show()
	end
end

local function Round(num, idp)
	local mult = 10^(idp or 0)
	return math.floor(num * mult + 0.5) / mult
end

function HermesDisplay:SecondsToBlizz(sSeconds)
	local nSeconds = tonumber(sSeconds)
	if nSeconds == 0 then
		return "0s";
	else
		hours = math.ceil(nSeconds/3600)
		mins = math.ceil(nSeconds/60)
		secs = math.ceil(nSeconds);
		
		if(nSeconds > 3600) then
			return tostring(hours).."h"
		end
		
		if(nSeconds > 60) then
			return tostring(mins).."m"
		end
		
		return tostring(math.floor(nSeconds)).."s"
	end
end

function HermesDisplay:SecondsToClock(sSeconds)
	local nSeconds = tonumber(sSeconds)
	if nSeconds == 0 then
		return "0";
	else
		nHours = string.format("%02.f", math.floor(nSeconds/3600));
		nMins = string.format("%02.f", math.floor(nSeconds/60 - (nHours*60)));
		nSecs = string.format("%02.f", math.floor(nSeconds - nHours*3600 - nMins *60));
		
		if(nSeconds > 3600) then
			return nHours..":"..nMins..":"..nSecs
		end
		
		if(nSeconds > 60) then
			return nMins..":"..nSecs
		end
		
		return Round(nSeconds)
	end
end

--------------------------------------------------------------------
-- Helpers
--------------------------------------------------------------------
function RotateTexture(self, angle)
	local function GetCorner(angle)
		local Root2 = 2 ^ 0.5;
		return 0.5 + cos( angle ) / Root2, 0.5 + sin( angle ) / Root2;
	end

	local LRx, LRy = GetCorner( angle + 45 );
	local LLx, LLy = GetCorner( angle + 135 );
	local ULx, ULy = GetCorner( angle - 135 );
	local URx, URy = GetCorner( angle - 45 );
	self:SetTexCoord( ULx, ULy, LLx, LLy, URx, URy, LRx, LRy );
end

