local AddonName, Hermes = ...

local LIB_AceAddon = LibStub("AceAddon-3.0") or error("Required library AceAddon-3.0 not found")
_G[AddonName] = LIB_AceAddon:NewAddon(Hermes, AddonName, "AceConsole-3.0", "AceEvent-3.0", "AceTimer-3.0", "AceComm-3.0", "AceSerializer-3.0")
local L = LibStub('AceLocale-3.0'):GetLocale('Hermes')

local dbp

--------------------------------------------------------------------
-- Static Data Structures
--------------------------------------------------------------------
--local independent class lookup, used mostly to keep message size down when sending messages.
Hermes.CLASS_ENUM = {
	[1]		= "ANY",
	[2]		= "DEATHKNIGHT",
	[3]		= "DRUID",
	[4]		= "HUNTER",
	[5]		= "MAGE",
	[6]		= "PALADIN",
	[7]		= "PRIEST",
	[8]		= "ROGUE",
	[9]		= "SHAMAN",
	[10]	= "WARLOCK",
	[11]	= "WARRIOR",
}

local MESSAGE_ENUM = {
	[1]		= "INITIALIZE_SENDER",
	[2]		= "INITIALIZE_RECEIVER",
	[3]		= "REQUEST_SPELLS",
	[4]		= "UPDATE_SPELLS",
	--[5]		= "REQUEST_SPELL_UPDATE",
}

local HermesDisplay = nil

local HERMES_SEND_COMM = "Hermes_Send_V1"
local HERMES_RECEIVE_COMM = "Hermes_Receive_V1"

Hermes.PlayerName = nil
Hermes.PlayerClass = nil
Hermes.PlayerInRaid = false
Hermes.PlayerInParty = false

Hermes.Sender = nil
Hermes.Receiver = nil

local CooldownScanTimer = nil
local COOLDOWN_SCAN_FREQUENCY = 1
local COOLDOWN_DELTA_THRESHOLD = COOLDOWN_SCAN_FREQUENCY * 5	--used to avoid sending messages for cooldowns that have expired normally
																--and used to determine whether cooldowns have decreased unexpectedly
																--(such as from tier set bonusus or anything else that might reduce a cooldown)

local localizedClassNames = {}
FillLocalizedClassList(localizedClassNames)

local defaultSpells = {
	{"PALADIN",		31821,	nil},					--Aura Mastery
	{"SHAMAN",		2825,	"Horde"},				--Bloodlust
	{"WARLOCK",		693,	nil},					--Create Soulstone
	{"PALADIN",		70940,	nil},					--Divine Guardian
	--{"PRIEST",		64843,	nil},					--Divine Hymn
	{"PRIEST",		47788,	nil},					--Guardian Spirit
	{"PALADIN",		6940,	nil},					--Hand of Sacrifice
	{"SHAMAN",		32182,	"Alliance"},			--Heroism
	--{"DRUID",		29166,	nil},					--Innervate
	{"PALADIN",		633,	nil},					--Lay on Hands
	{"PRIEST",		33206,	nil},					--Pain Suppression
	{"DRUID",		20484,	nil},					--Rebirth
	--{"SHAMAN",		20608,	nil},					--Reincarnation
}
--------------------------------------------------------------------
-- WoW Events
--------------------------------------------------------------------
local PLAYER_ENTERED_WORLD = nil
function Hermes:PLAYER_ENTERING_WORLD()				--used only for one time player initialization that has to happen after player is logged in.
	self:UnregisterEvent("PLAYER_ENTERING_WORLD")
	
	--initialize static properties for player
	self.PlayerName = UnitName("player")
	self.PlayerClass = Hermes:GetClassEnum(select(2, UnitClass("player")))
	PLAYER_ENTERED_WORLD = true
	self:UpdateCommunicationsStatus()
end

function Hermes:UpdateCommunicationsStatus()
	local wasInRaid = self.PlayerInRaid
	local wasInParty = self.PlayerInParty
	
	if(GetNumRaidMembers() > 0) then
		--we're in a raid
		self.PlayerInRaid = true
	else
		self.PlayerInRaid = false
	end
	
	if(GetNumPartyMembers() > 0) then
		--we're in a party (or raid)
		self.PlayerInParty = true
	else
		self.PlayerInParty = false
	end
	
	--this is a special case required so that whenever we go from a party to a raid, or vice verse, that we reset sending and receiving.
	--we can just stop it here if that's the case, and allow the code below to restart it if necessary
	--if((wasInRaid == true and self.PlayerInRaid == false) or (wasInParty == true and self.PlayerInParty == false)) then
	if(	(wasInParty == true and wasInRaid == false and self.PlayerInRaid == true) 			--we just converted from a party to a raid
		or 
		(wasInRaid == true and self.PlayerInRaid == false and self.PlayerInParty == true) ) then	--we just converted from a raid to a party
		
		if(self:IsSending() == true ) then
			self:StopSending()
		end
		if(self:IsReceiving() == true ) then
			self:StopReceiving()
		end
	end
	
	--start sending if needed
	if(dbp.enabled == true and dbp.sender.enabled == true and ((self.PlayerInParty == true and dbp.enableparty == true) or self.PlayerInRaid == true)) then
		if(self:IsSending() == false) then
			self:StartSending()
		end
	else
		if(self:IsSending() == true ) then
			self:StopSending()
		end
	end
	
	--start receiving if needed
	if(dbp.enabled == true and dbp.receiver.enabled == true and ((self.PlayerInParty == true and dbp.enableparty == true) or self.PlayerInRaid == true)) then
		if(self:IsReceiving() == false) then
			self:StartReceiving()
		end
	else
		if(self:IsReceiving() == true ) then
			self:StopReceiving()
		end
	end
end

function Hermes:RAID_ROSTER_UPDATE()
	self:UpdateCommunicationsStatus()
end

function Hermes:PARTY_MEMBERS_CHANGED()
	self:UpdateCommunicationsStatus()
end

function Hermes:ACTIVE_TALENT_GROUP_CHANGED()
	--reset sender
	Hermes:StopSending()
	Hermes:StartSending()
end

function Hermes:ShowHideHermesDisplay()
	local show = false								--default to hide
	if dbp.display.autohide == false then		--always show it if this is set
		show = true
	elseif self:IsReceiving() == true then					--show if we're currently receiving
		show = true
	end
	
	HermesDisplay:Show(show)
end

--------------------------------------------------------------------
-- Receiver
--------------------------------------------------------------------
function Hermes:StartReceiving()
	--init Receiver
	Hermes.Receiver = {}

	HermesDisplay:Startup()
	--HermesDisplay:EnableEvents(true)
	Hermes:ShowHideHermesDisplay()
	
	--be ready for messages before sending requests
	self:RegisterComm(HERMES_RECEIVE_COMM, "OnReceiverComm")

	--advertise receiver
	self:SendMessage_INITIALIZE_RECEIVER(nil)
end

function Hermes:StopReceiving()
	HermesDisplay:Shutdown()
	
	--cleanup receiver state
	if(Hermes.Receiver) then
		wipe(Hermes.Receiver)
		Hermes.Receiver = nil
	end
	
	--have to call this after we set receiver to nil so that IsReceiving returns false
	Hermes:ShowHideHermesDisplay()
	
	--stop getting messages
	self:UnregisterComm(HERMES_RECEIVE_COMM)
end

function Hermes:RequestSpellUpdate(spellId)
	local spellRequests = {}
	tinsert(spellRequests, spellId)
	self:SendMessage_REQUEST_SPELLS(nil, spellRequests)
end

function Hermes:HandleSender(senderName, classEnum, resetIfExists)
	--see if we know the sender
	local sender = HermesDisplay:GetSender(senderName)
	if(sender) then
		if(not resetIfExists) then
			return
		else
			HermesDisplay:RemoveSpellsForSender(senderName)
		end
	end
	
	if(not sender) then
		sender = HermesDisplay:AddSender(senderName, classEnum)
	end
	
	--build up spellRequests
	local spellRequests = {}
	for i, spell in ipairs(dbp.receiver.spells) do
		--only add spells that apply to the class
		if(Hermes:GetClassEnum(spell.class) == classEnum or spell.class == "ANY") then
			tinsert(spellRequests, spell.spellid)
		end
	end
	
	--only send message if there are spells requested for the senders class
	if(#spellRequests > 0) then
		self:SendMessage_REQUEST_SPELLS(senderName, spellRequests)
	end
end

function Hermes:ProcessMessage_INITIALIZE_SENDER(senderName, classEnum, channel)
	self:HandleSender(senderName, classEnum, channel ~= "WHISPER")
end

function Hermes:HandleSpellUpdates(senderName, classEnum, spellUpdates)
	self:HandleSender(senderName, classEnum, false) --adds the sender if it wasn't already known

	for _, spellUpdate in ipairs(spellUpdates) do
		local spell = nil
		for i, s in ipairs(dbp.receiver.spells) do
			if(s.spellid == spellUpdate[1] and (Hermes:GetClassEnum(s.class) == classEnum or s.class == "ANY")) then
				spell = s --we're tracking it
				break
			end
		end
		
		--only proceed if we're tracking this spell
		if(spell) then
			--now update it
			local duration = spellUpdate[2]
			HermesDisplay:AddSenderSpellInfo(senderName, spell.spellid, duration)
		end
	end
end

function Hermes:ProcessMessage_UPDATE_SPELLS(senderName, classEnum, spellUpdates, channel)
	local spell = self:HandleSpellUpdates(senderName, classEnum, spellUpdates)
end

function Hermes:IsReceiving()
	return Hermes.Receiver ~= nil
end

--------------------------------------------------------------------
-- Sender
--------------------------------------------------------------------
function Hermes:StartSending()
	--init Sender
	Hermes.Sender = {}
	Hermes.Sender.spells = {}

	--be ready for messages before start sending
	self:RegisterComm(HERMES_SEND_COMM, "OnSenderComm")

	--register for other wow events
	self:RegisterEvent("ACTIVE_TALENT_GROUP_CHANGED")

	--advertise self to receivers
	self:SendMessage_INITIALIZE_SENDER(nil)

	self:StartCooldownScanTimer()
end

function Hermes:StopSending()
	--cleanup sender state
	if(Hermes.Sender) then
		wipe(Hermes.Sender)
		Hermes.Sender = nil
	end
	
	--stop getting messages
	self:UnregisterComm(HERMES_SEND_COMM)
	
	--unregister for other wow events
	self:UnregisterEvent("ACTIVE_TALENT_GROUP_CHANGED")
	
	self:KillCooldownScanTimer()
end

function Hermes:HandleSpellRequests(receiverName, spellRequests)
	--see if the class applies to us and that we know the spell
	for _, spellid in ipairs(spellRequests) do
		if(Hermes:IsSpellKnown(spellid)) then
			local spell = nil
			--see if we know the spell
			for i, s in ipairs(Hermes.Sender.spells) do
				if(s.spellid == spellid) then
					--we know the spell, see if we know the receiver
					local receiver = s.receivers[receiverName]
					--we don't know this receiver yet, add it
					if(not receiver) then
						s.receivers[receiverName] = {}
					end
					s.force = true --force the spell to broadcast on next update
					spell = s --mark spell for below...
					break
				end
			end
			
			--we're not tracking this spell yet
			if(not spell) then
				spell = {
					spellid = spellid,
					duration = nil,
					force = true, --force the spell to send on next update
					receivers = {}
				}
				spell.receivers[receiverName] = {} --add the receiver to the spell
				tinsert(Hermes.Sender.spells, spell) --add the spell
			end
		end
	end
end

function Hermes:ProcessMessage_REQUEST_SPELLS(receiverName, spellRequests, channel)
	self:HandleSpellRequests(receiverName, spellRequests)
end

function Hermes:ProcessMessage_INITIALIZE_RECEIVER(receiverName, channel)
	self:SendMessage_INITIALIZE_SENDER(receiverName)
end

function Hermes:IsSending()
	return Hermes.Sender ~= nil
end

--------------------------------------------------------------------
-- AceAddon
--------------------------------------------------------------------
function Hermes:OnEnable()
	HermesDisplay = self:GetModule("HermesDisplay")
	self:Startup()
end

function Hermes:OnDisable()
	self:Shutdown()
end

function Hermes:OnInitialize()
	self:LoadDefaultOptions()
	dbp = self.db.profile
	self:LoadBlizOptions()

	self.db.RegisterCallback(self, "OnNewProfile", "OnProfileChanged")
	self.db.RegisterCallback(self, "OnProfileChanged", "OnProfileChanged")
	self.db.RegisterCallback(self, "OnProfileCopied", "OnProfileChanged")
	self.db.RegisterCallback(self, "OnProfileReset", "OnProfileChanged")
end

function Hermes:Shutdown()
	--unregister events
	self:UnregisterEvent("RAID_ROSTER_UPDATE")
	self:UnregisterEvent("PARTY_MEMBERS_CHANGED")
	self:StopSending()
	self:StopReceiving()
	
	HermesDisplay:ResetAll()
	HermesDisplay:Show(false) --ensure the display is hidden
end

function Hermes:Startup()
	HermesDisplay:ResetAll() --make sure that the display is initialized before starting up comm messages
	
	--register events
	if(PLAYER_ENTERED_WORLD) then
		self:UpdateCommunicationsStatus()
	else
		self:RegisterEvent("PLAYER_ENTERING_WORLD")
	end
	
	self:RegisterEvent("RAID_ROSTER_UPDATE")
	self:RegisterEvent("PARTY_MEMBERS_CHANGED")
	
	if(not dbp.welcome)	then
		Hermes:ShowWelcomeMessage()
	end

	--build up the spell tracking list
	for i, s in ipairs(dbp.receiver.spells) do
		HermesDisplay:AddSpell(s.spellid, s.class)
	end
	
	Hermes:ShowHideHermesDisplay()
end

function Hermes:ShowWelcomeMessage()
	local message = L["Message from Hermes..."] .. "\n\n"
	message = message ..L["|cffD9D919Greetings young hero. Zeus recommends I supply new travelers with the most valued of spells.|r"].."\n\n"
	message = message ..L["|cff00ff00How say you?|r"]
	StaticPopupDialogs["HermesWelcomeMessage"] = {
		text = message, 
		button1 = L["Sure!"], 
		button2 = L["Naw, I'm good"],
		timeout = 0,
		whileDead = 1, 
		hideOnEscape = 1, 
		OnAccept = function() Hermes:SetupDefaultSpells() end,
		--OnAlt = function() print("Shoes") end,
		OnCancel = function() end
	}
	StaticPopup_Show("HermesWelcomeMessage")
	dbp.welcome = true
end

function Hermes:SetupDefaultSpells()
	local playerAlliance = select(1, UnitFactionGroup("player"))
	
	for i, default in ipairs(defaultSpells) do
		local class = default[1]
		local spellid = default[2]
		local alliance = default[3]
		
		if(not alliance or (alliance and alliance == playerAlliance)) then
			--make sure spell isn't already being tracked
			local exists = false
			for _, s in ipairs(dbp.receiver.spells) do
				if(s.spellid == spellid) then
					exists = true
					break
				end
			end
			
			if(exists == false) then
				local spell = Hermes:FindSpellName(class, spellid)
				if(spell) then
					--add to db
					tinsert(dbp.receiver.spells, spell)
					--sort the spells
					table.sort(dbp.receiver.spells, function(a, b) return Hermes:SortProfileSpells(a, b) end)

					local HermesDisplay = self:GetModule("HermesDisplay")
					HermesDisplay:AddSpell(spell.spellid, spell.class)
					
					if(Hermes:IsReceiving()) then
						--request spell info from senders
						self:SendMessage_REQUEST_SPELLS(nil, {spell.spellid})
					end
				end
			end
		end
	end
	
	--update blizzard options
	Hermes:BlizOptionsTable_Spells()
end

function Hermes:FindSpellName(class, spellid)
	--see if the spell exists
	local name
	local icon
	name, _, icon, _, _, _, _, _, _ = GetSpellInfo(spellid)
	
	if(not name or not icon) then
		return nil
	end
	
	local spell = {
		class = class,
		spellid = spellid,
		spellname = name,
		spellicon = icon,
	}

	return spell
end

function Hermes:OnProfileChanged()
	dbp = self.db.profile
	self:Disable()
	self:UpdateBlizOptionsTableReferences() --update table references, otherwise they'll be pointing to old tables for the prior profile
	self:Enable() 
	HermesDisplay:RestoreLayoutFramePosition()
end
--------------------------------------------------------------------
-- Message Sending
--------------------------------------------------------------------
function Hermes:SendMessageToReceivers(recipientName, msgEnum, msgContent)
	local msg = self:Serialize({msgEnum, msgContent})
	local channel = self:GetAppropriateMessageChannel(recipientName)
	self:SendCommMessage(HERMES_RECEIVE_COMM, msg, channel, recipientName, "NORMAL", nil, nil)
end

function Hermes:GetAppropriateMessageChannel(recipientName)
	if(recipientName) then
		return "WHISPER"
	elseif(self.PlayerInRaid == true) then
		return "RAID"
	elseif(self.PlayerInParty == true) then
		return "PARTY"
	else
		error("Unable to determine message channel")
	end
end

function Hermes:SendMessageToSenders(recipientName, msgEnum, msgContent)
	local msg = self:Serialize({msgEnum, msgContent})
	local channel = self:GetAppropriateMessageChannel(recipientName)
	self:SendCommMessage(HERMES_SEND_COMM, msg, channel, recipientName, "NORMAL", nil, nil)
end

function Hermes:SendMessage_REQUEST_SPELLS(recipientName, spellRequests)
	self:SendMessageToSenders(recipientName, Hermes:GetMessageEnum("REQUEST_SPELLS"),
		{
			spellRequests
		}
	)
end

function Hermes:SendMessage_INITIALIZE_SENDER(recipientName)
	self:SendMessageToReceivers(recipientName, Hermes:GetMessageEnum("INITIALIZE_SENDER"),
		self.PlayerClass
	)
end

function Hermes:SendMessage_INITIALIZE_RECEIVER(recipientName)
	self:SendMessageToSenders(recipientName, Hermes:GetMessageEnum("INITIALIZE_RECEIVER"),
		nil
	)
end

function Hermes:SendMessage_UPDATE_SPELLS(recipientName, spellUpdates)
	self:SendMessageToReceivers(recipientName, Hermes:GetMessageEnum("UPDATE_SPELLS"),
		{
			self.PlayerClass,
			spellUpdates,
		}
	)
end

--------------------------------------------------------------------
-- Message Receiving
--------------------------------------------------------------------
function Hermes:OnReceiverComm(prefix, serialized, channel, sender)
	if(prefix == HERMES_RECEIVE_COMM) then
		local success, msg = self:Deserialize(serialized)
		if(success) then
			local msgEnum = msg[1]
			local msgContent = msg[2]
			local msgName = MESSAGE_ENUM[msgEnum]
			
			if(msgName and msgName == "INITIALIZE_SENDER") then
				local classEnum = msgContent
				self:ProcessMessage_INITIALIZE_SENDER(sender, classEnum, channel)
				
			elseif(msgName and msgName == "UPDATE_SPELLS") then
				local classEnum = msgContent[1]
				local spellUpdates = msgContent[2]
				self:ProcessMessage_UPDATE_SPELLS(sender, classEnum, spellUpdates, channel)
			end
		else
			error("Error deserializing message")
		end
	end
end

function Hermes:OnSenderComm(prefix, serialized, channel, sender)
	if(prefix == HERMES_SEND_COMM) then
		local success, msg = self:Deserialize(serialized)
		if(success) then
			local msgEnum = msg[1]
			local msgContent = msg[2]
			local msgName = MESSAGE_ENUM[msgEnum]
			
			if(msgName and msgName == "REQUEST_SPELLS") then
				local spellRequests = msgContent[1]
				self:ProcessMessage_REQUEST_SPELLS(sender, spellRequests, channel)

			elseif(msgName and msgName == "INITIALIZE_RECEIVER") then
				self:ProcessMessage_INITIALIZE_RECEIVER(sender, channel)
			end
		else
			error("Error deserializing message")
		end
	end
end

--------------------------------------------------------------------
-- Spell Tracking
--------------------------------------------------------------------
function Hermes:IsSpellKnown(spellid)
	local slot = FindSpellBookSlotBySpellID(spellid)
	if slot and slot <= MAX_SPELLS then
		local slotType, id = GetSpellBookItemInfo(slot, BOOKTYPE_SPELL)
		return id == spellid and slotType ~= 'FUTURESPELL'
	end
end

function Hermes:UpdateSenderCooldown(spell)
	local send = false
		
	if(spell.force == true) then
		send = true
		spell.force = false
	else
		local priorCooldown = spell.duration
		spell.duration = self:GetLatestCooldown(spell.spellid)
		
		--TESTING COOLDOWN CHANGE DETECTION
		--if(spell.duration and spell.duration < 30 ) then spell.duration = 5 end			--Force a DECREASE detection
		--if(spell.duration and spell.duration < 30 ) then spell.duration = 60 end			--Force a INCREASE detection
		--if(spell.duration and spell.duration < 30 ) then spell.duration = nil end			--Force a COOLDOWN reset detection
	
		if(priorCooldown ~= spell.duration) then
			--if it wasn't on cooldown last but it is now
			if(priorCooldown == nil and spell.duration ~= nil) then
				send = true
				
			--if it was on cooldown but now it's not.
			elseif(priorCooldown ~= nil and spell.duration == nil) then
				--only send message if the last cooldown value is greater than the threshold, this is to prevent sending messages for spells that went off cooldown naturally
				if(priorCooldown > COOLDOWN_DELTA_THRESHOLD) then
					send = true
				end
				
			elseif(priorCooldown ~= nil and spell.duration ~= nil) then
				--if for some reason the cooldown is larger than it was last time we checked
				if(spell.duration > priorCooldown) then
					send = true
				end
				--if for some reason the cooldown is unexpectedly smaller than last time we checked
				if((priorCooldown - spell.duration) > COOLDOWN_DELTA_THRESHOLD) then
					send = true
				end
			end
		end
	end

	return send
end

function Hermes:GetLatestCooldown(spellid)
	local start, duration, _ = GetSpellCooldown(spellid)
	--note that GetSpellCooldown returns the GCD if active when called. By calling > 2 I hope to filter out any GCD's at the expense of not reporting cooldowns actually only having 2 seconds or less on them.
	if(duration > 2) then
		duration = ceil(duration - (GetTime() - start)) --spell on cooldown
	else
		duration = nil
	end
	
	return duration
end

--------------------------------------------------------------------
-- Helpers
--------------------------------------------------------------------
function Hermes:GetClassEnum(classFileName)
	for i, class in ipairs(Hermes.CLASS_ENUM) do
		if(class == classFileName) then
			return i
		end
	end
	error("Failed to locate class enum")
end

function Hermes:GetMessageEnum(messageName)
	for i, msg in ipairs(MESSAGE_ENUM) do
		if(msg == messageName) then
			return i
		end
	end
	error("Failed to locate message enum")
end

function Hermes:CloneTable(tbl, cloneV, cloneK)
	assert(not flag[tbl], "CloneTable(): cannot deep-clone a table that contains a reference to itself")
	flag[tbl] = 1
	local newtbl = {}
	for k,v in pairs(tbl) do
		if (cloneK and type(k)=="table") then k = CloneTable(k, cloneV, cloneK) end
		if (cloneV and type(v)=="table") then v = CloneTable(v, cloneV, cloneK) end
		newtbl[k] = v
	end
	flag[tbl] = nil
	return newtbl
end

function Hermes:KillCooldownScanTimer()
	if(CooldownScanTimer) then
		self:CancelTimer(CooldownScanTimer, true)
		CooldownScanTimer = self:ScheduleTimer("OnUpdateSenderCooldowns", 0.1, nil) --convert to one shot timer
		self:CancelTimer(CooldownScanTimer, true) --cancel again, we don't care if it ran or not
		CooldownScanTimer = nil
	end
end

function Hermes:StartCooldownScanTimer()
	CooldownScanTimer = self:ScheduleRepeatingTimer("OnUpdateSenderCooldowns", COOLDOWN_SCAN_FREQUENCY, nil)
end

function Hermes:OnUpdateSenderCooldowns()
	if(Hermes.Sender and Hermes.Sender.spells) then
		--first update all of the cooldowns and populate list with spells that need to be sent
		local spellUpdates = nil
		for i, spell in ipairs(Hermes.Sender.spells) do
			if(self:UpdateSenderCooldown(spell) == true) then
				if(not spellUpdates) then
					spellUpdates = {}
				end
				tinsert(spellUpdates, {spell.spellid, spell.duration})
			end
		end
		
		--now determine if any messages need to be sent
		if(spellUpdates and #spellUpdates > 0) then
			self:SendMessage_UPDATE_SPELLS(nil, spellUpdates)
		end
	end
end

function Hermes:GetLocalizedClassName(classToken)
	local localizedClassName
	if(classToken == "ANY") then
		return L["Any"]
	else
		return localizedClassNames[classToken]
	end
end

function Hermes:GetClassColorRGB(class)
	local number = tonumber(class)
	if(number) then --they used the enum number
		local classColorRGB
		if(class == 1) then
			classColorRGB = RAID_CLASS_COLORS["PRIEST"] --white
		else
			classColorRGB = RAID_CLASS_COLORS[Hermes.CLASS_ENUM[class]]
		end
		return classColorRGB
	else --they used the enum value
		local classColorRGB
		if(class == "ANY") then
			classColorRGB = RAID_CLASS_COLORS["PRIEST"] --white
		else
			classColorRGB = RAID_CLASS_COLORS[class]
		end
		return classColorRGB
	end
end

function Hermes:GetClassColorHEX(class)
	local number = tonumber(class)
	if(number) then --they used the enum number
		local classColorRGB
		if(class == 1) then
			classColorRGB = RAID_CLASS_COLORS["PRIEST"] --white
		else
			classColorRGB = RAID_CLASS_COLORS[Hermes.CLASS_ENUM[class]]
		end
		local classColorHex = string.format("FF%02x%02x%02x", 
			classColorRGB.r * 255,
			classColorRGB.g * 255,
			classColorRGB.b * 255)
		return classColorHex
	else --they used the enum value
		local classColorRGB
		if(class == "ANY") then
			classColorRGB = RAID_CLASS_COLORS["PRIEST"] --white
		else
			classColorRGB = RAID_CLASS_COLORS[class]
		end
		local classColorHex = string.format("FF%02x%02x%02x", 
			classColorRGB.r * 255,
			classColorRGB.g * 255,
			classColorRGB.b * 255)
		return classColorHex
	end
end

function Hermes:GetColorCodedClassText(text, class)
	return "|c"..self:GetClassColorHEX(class)..text.."|r"
end

function Hermes:Trim (s)
	return (string.gsub(s, "^%s*(.-)%s*$", "%1"))
end

function Hermes:SortProfileSpells(a, b)
	return a.spellname < b.spellname
end

function Hermes:GetSpellID(spell)
	local matches = {}
	for i = 1, 100000 do
		name, rank, icon, powerCost, _, powerType, castingTime, minRange, maxRange = GetSpellInfo(i)
		if name == spell then
			tinsert(matches, i)
		end
	end
	
	return matches
end


