local AddonName, Hermes = ...

local LIB_AceConfig = LibStub("AceConfig-3.0") or error("Required library AceConfig-3.0 not found")
local LIB_AceDB = LibStub("AceDB-3.0") or error("Required library AceDB-3.0 not found")
local LIB_AceConfigDialog = LibStub("AceConfigDialog-3.0")  or error("Required library AceConfigDialog-3.0 not found")
local LIB_AceConfigRegistry = LibStub("AceConfigRegistry-3.0") or error("Required library AceConfigRegistry-3.0 not found")
local L = LibStub('AceLocale-3.0'):GetLocale('Hermes')

local newSpellTemplate = {
	class = L["-- Select --"],
	spellid = nil,
	spellname = nil,
}

function Hermes:LoadDefaultOptions()
	local defaults = {
		profile = {
			enabled = true,
			enableparty = false,
			sender = {
				enabled = true,
			},
			receiver = {
				enabled = true,
				spells = {},
			},
			display = {
				locked = false,
				autohide = false,
				scale = 1,
				padding = 5,
				buttonanchor = "TOPLEFT",
				lbf = {
					skin = "Blizzard",
					gloss = 0,
					backdrop = false,
					colors = {},
				},
				VertexColorUnavailable = {0.5, 0.5, 0.5, 1.0},
				VertexColorDoesntExist = {1.0, 1.0, 1.0, 0.5},
			}
		},
	}
	self.db = LIB_AceDB:New("HermesDB", defaults)
end

function Hermes:LoadBlizOptions()
	optionTables = {
		handler = Hermes,
		type="group",
		childGroups = "tab",
		name = "Hermes",
		args = {
			General = {
				name = L["General"], type="group", order = 0,
				args = {
					General = {
						type = "group",
						name = L["General"],
						inline = true,
						order = 0,
						args = {
							EnableHermes = {
								type = "toggle",
								name = L["Enable Hermes"],
								width = "full",
								get = function(info) return self.db.profile.enabled end,
								order = 0,
								set = "OnSetEnabled",
							},
						},
					},
					Communication = {
						type = "group",
						name = L["Communication"],
						inline = true,
						order = 1,
						disabled = function() return self.db.profile.enabled == false end,
						args = {
							EnableSender = {
								type = "toggle",
								name = L["Enable Sending"],
								width = "normal",
								get = function(info) return Hermes.db.profile.sender.enabled end,
								order = 2,
								set = "OnSetEnableSender",
							},
							Spacer2A = {
								type = "description",
								name = L["Send cooldown messages to other players. No additional configuration is required for sending."],
								width = "double",
								order = 3,
							},
							Spacer2B = {
								type = "description",
								name = "",
								width = "full",
								order = 4,
							},
							EnableReceiver = {
								type = "toggle",
								name = L["Enable Receiving"],
								width = "normal",
								get = function(info) return Hermes.db.profile.receiver.enabled end,
								order = 5,
								set = "OnSetEnableReceiver",
							},
							Spacer3A = {
								type = "description",
								name = L["Receive cooldown messages from other players. The spells configured in the Spells tab will be requested from all Hermes users with Sending enabled."],
								width = "double",
								order = 6,
							},
							Spacer3B = {
								type = "description",
								name = "",
								width = "full",
								order = 7,
							},
							EnableParty = {
								type = "toggle",
								name = L["Enable Party Support"],
								get = function(info) return Hermes.db.profile.enableparty end,
								order = 8,
								set = "OnSetEnablePartySupport",
							},
							Spacer4A = {
								type = "description",
								name = L["By default, Hermes only communicates during raids. Check this box to allow party communications as well."],
								width = "double",
								order = 9,
							},
							Spacer4B = {
								type = "description",
								name = "",
								width = "full",
								order = 10,
							},
						},
					},
				},
			},
			Spells = {
				name=L["Spells"], type="group", childGroups="tab", order = 1,
				disabled = function() return self.db.profile.enabled == false end,
				args = { }
			},
			Window = {
				name = L["Window"], type="group", order = 2,
				disabled = function() return self.db.profile.enabled == false end,
				args = {
					General = {
						type = "group",
						name = L["General"],
						inline = true,
						order = 0,
						args = {
							LockWindow = {
								type = "toggle",
								name = L["Lock Window"],
								width = "full",
								get = function(info) return Hermes.db.profile.display.locked end,
								order = 0,
								set = "OnSetLockDisplay",
							},
							AutoHide = {
								type = "toggle",
								name = L["Auto Hide"],
								width = "full",
								desc = L["Hides the window when not Receiving."],
								get = function(info) return Hermes.db.profile.display.autohide end,
								order = 1,
								set = "OnSetAutoHide",
							},
							ButtonAnchor = {
								type = "select",
								name = L["Button Anchor"],
								width = "full",
								desc = "Changes the direction in which buttons grow.",
								get = function(info) return Hermes.db.profile.display.buttonanchor end,
								order = 2,
								set = "OnSetButtonAnchor",
								values = {
									["TOPLEFT"] = L["Top Left"],
									["TOPRIGHT"] = L["Top Right"],
									["BOTTOMLEFT"] = L["Bottom Left"],
									["BOTTOMRIGHT"] = L["Bottom Right"],
								},
							},
							Scale = {
								type = "range",
								min = 0.1, max = 3, step = 0.01,
								name = L["Scale"],
								order = 3,
								width = "full",
								get = function(info) return Hermes.db.profile.display.scale end,
								set = function(info, value)
									Hermes.db.profile.display.scale = value
									local HermesDisplay = self:GetModule("HermesDisplay")
									HermesDisplay:SetWindowScale(value)
								end,
							},
							Padding = {
								type = "range", min = 0, max = 100, step = 1,
								name = L["Padding"],
								width = "full",
								get = function(info) return Hermes.db.profile.display.padding end,
								order = 4,
								set = "OnSetPadding",
							},
						},
					},
				},
			},
		},
	}
	
	self:UpdateBlizOptionsTableReferences()
	LIB_AceConfig:RegisterOptionsTable(AddonName, optionTables)
	self.blizzOptionsFrame = LIB_AceConfigDialog:AddToBlizOptions(AddonName, AddonName)
	optionTables.args.Profiles = LibStub("AceDBOptions-3.0"):GetOptionsTable(self.db)
	optionTables.args.Profiles.order = 100
	LIB_AceConfigDialog:SetDefaultSize(AddonName, 700, 550)
end

function Hermes:UpdateBlizOptionsTableReferences()
	self:BlizOptionsTable_Spells()
end

function Hermes:BlizOptionsTable_Spells()
	if(optionTables.args.Spells.args) then
		wipe(optionTables.args.Spells.args)
	end

	--setup basic class structure
	for classEnum, className in ipairs(Hermes.CLASS_ENUM) do
		
		local classGroup = {
			type = "group",
			inline = true,
			name = Hermes:GetColorCodedClassText(Hermes:GetLocalizedClassName(className), className),
			order = classEnum,
			args = { }
		}
		
		local groupHasSpells = false
		for i, spell in ipairs(self.db.profile.receiver.spells) do
			if(spell.class == className) then
				groupHasSpells = true
				classGroup.args[tostring(spell.spellid)] = {
					type = "group",
					inline = true,
					name = "",
					order = i,
					args = {
						spellname = {
							type = "description",
							name = "|T"..spell.spellicon..":0:0:0:0|t "..spell.spellname,
							order = 0,
							width = "normal",
						},
						spellid = {
							type = "description",
							name = tostring(spell.spellid),
							order = 1,
							width = "normal",
						},
						DeleteButton = {
							type = "execute",
							name = L["Delete"],
							width = "normal",
							order = 2,
							func = "OnDeleteSpell",
						}
					}
				}
			end
		end
		
		if(groupHasSpells) then
			optionTables.args.Spells.args[className] = classGroup
		end
	end
	
	optionTables.args.Spells.args.Create = {
		type="group",
		name=L["Add Spell"],
		inline = true,
		order = 0,
		args = {
			ClassSelect = {
				type = "select",
				name = L["Class"],
				order = 0,
				style = "radio",
				values = function()
					local values = {}
					values[L["-- Select --"]] = L["-- Select --"]
					for i, classFileName in ipairs(Hermes.CLASS_ENUM) do
						values[classFileName] = Hermes:GetColorCodedClassText(Hermes:GetLocalizedClassName(classFileName), classFileName)
					end
					return values
				end,
				get = function(info) return newSpellTemplate.class end,
				set = function(info, value) newSpellTemplate.class = value end,
			},
			SpellNameOrId = {
				type = "input",
				name = L["Spell Name or ID"],
				order = 1,
				get = function(info)
					if(newSpellTemplate.spellname ~= nil) then
						return newSpellTemplate.spellname
					elseif(newSpellTemplate.spellid ~= nil) then
						return string.format("%.0f", newSpellTemplate.spellid)
					else
						return ""
					end
				end,
				set = function(info, value)
					local n = tonumber(self:Trim(value))
					local s = self:Trim(value)
					if(strlen(s) == 0) then s = nil end
					if(n and n ~= 0) then
						newSpellTemplate.spellid = n
						newSpellTemplate.spellname = nil
					else
						newSpellTemplate.spellid = nil
						newSpellTemplate.spellname = s
					end
				end,
			},
			AddButton = {
				type = "execute",
				name = L["Add"],
				width = "normal",
				order = 3,
				func = "OnAddSpell",
				disabled = function()
					--ensure that a class is selected and that spell evaluates to a spell id or name
					return newSpellTemplate.class == L["-- Select --"] or (not newSpellTemplate.spellname and not newSpellTemplate.spellid)
				end,
			}
		}
	}
end

function Hermes:OnSetEnabled(info, value)
	self.db.profile.enabled = value
	if(value == true) then
		Hermes:Startup()
	else
		Hermes:Shutdown()
	end
end

function Hermes:OnSetEnablePartySupport(info, value)
	self.db.profile.enableparty = value
	self:UpdateCommunicationsStatus()
end

function Hermes:OnSetEnableAnimations(info, value)
	self.db.profile.display.enableanimations = value
end

function Hermes:OnSetEnableSender(info, value)
	self.db.profile.sender.enabled = value
	self:UpdateCommunicationsStatus()
end

function Hermes:OnSetEnableReceiver(info, value)
	self.db.profile.receiver.enabled = value
	self:UpdateCommunicationsStatus()
end

function Hermes:OnSetLockDisplay(info, value)
	self.db.profile.display.locked = value
	local HermesDisplay = self:GetModule("HermesDisplay")
	HermesDisplay:Lock(value)
end

function Hermes:OnSetAutoHide(info, value)
	self.db.profile.display.autohide = value
	self:ShowHideHermesDisplay()
end

function Hermes:OnSetPadding(info, value)
	self.db.profile.display.padding = value
	local HermesDisplay = self:GetModule("HermesDisplay")
	HermesDisplay:ArrangeLayout()
end

function Hermes:OnSetButtonAnchor(info, value)
	self.db.profile.display.buttonanchor = value
	local HermesDisplay = self:GetModule("HermesDisplay")
	HermesDisplay:UpdateFrameResizerPosition()
	--HermesDisplay:AnchorCommStatusFrame()
	HermesDisplay:ArrangeLayout()
end

function Hermes:OnAddSpell(info)
	--see if the spell exists
	local name
	local icon
	if(newSpellTemplate.spellid) then
		name, _, icon, _, _, _, _, _, _ = GetSpellInfo(newSpellTemplate.spellid)
	else
		--try to find the spell id for the given spell name
		local matches = self:GetSpellID(newSpellTemplate.spellname)
		
		if(matches and #matches > 0) then
			--still need the icon
			newSpellTemplate.spellid = matches[1]
			name, _, icon, _, _, _, _, _, _ = GetSpellInfo(newSpellTemplate.spellid)
			if(#matches > 1) then
				--there are more than one spellid from the name given, ask the user which they want to use
				local message = "\n"
				message = message .. L["|cFFFF0000Hermes Warning|r"].."\n"
				message = message .. string.format(L["%s spellid's were found for %s, the |cFF00FF00highlighted|r spellid was chosen:"] .. "\n",
					tostring(#matches),
					newSpellTemplate.spellname)

				for i = 1, #matches do
					--*** NOTE: Disabled querying/showing spell links for now because I don't have good knowledge of what it takes to get disconnected
					--local link = GetSpellLink(matches[i])
					if(i == 1) then
						message = message.."   |cFF00FF00"..tostring(matches[i]).."|r\n"
					else
						message = message.."   "..tostring(matches[i]).."\n"
					end
					--if(link) then
					--	message = message..link.."\n"
					--else
					--	message = message.."(no link available)\n"
					--end
				end
				
				message = message..L["If this is incorrect, delete the spell just created and use one of the spellids above."]
				print(message)
			end
		end
	end
	
	if(name and icon) then
		--make sure we're not already tracking this spell
		for i, spell in ipairs(self.db.profile.receiver.spells) do
			if(spell.spellid == newSpellTemplate.spellid) then
				local message = L["Message from Hermes: |cFFFF0000spell has already been added|r"]
				UIErrorsFrame:AddMessage(message)
				return
			end
		end
		
		local spell = {
			class = newSpellTemplate.class,
			spellid = newSpellTemplate.spellid,
			spellname = name,
			spellicon = icon,
		}
		
		--add to db
		tinsert(self.db.profile.receiver.spells, spell)
		
		--sort the spells
		table.sort(self.db.profile.receiver.spells, function(a, b) return Hermes:SortProfileSpells(a, b) end)

		--add to display if receiver is running
		--if(Hermes:IsReceiving()) then
		--add to display
		local HermesDisplay = self:GetModule("HermesDisplay")
		HermesDisplay:AddSpell(spell.spellid, spell.class)
		--end
		
		--reset the template, but leave the class alone
		newSpellTemplate.spellid = nil
		newSpellTemplate.spellname = nil
		
		--request spell from senders
		if(Hermes:IsReceiving()) then
			self:SendMessage_REQUEST_SPELLS(nil, {spell.spellid})
		end
	
		--force the config window to update with latest spell info
		LIB_AceConfigRegistry:NotifyChange(AddonName)
		Hermes:BlizOptionsTable_Spells()
	else
		local message = L["Message from Hermes: |cFFFF0000spell name or id not found|r"]
		UIErrorsFrame:AddMessage(message)
	end
end

function Hermes:OnDeleteSpell(info)
	local spellid = info[#info - 1]
	local classToken = info[#info - 2]
	
	--locate the index of the spell in the db
	local index = nil
	for i, spell in ipairs(self.db.profile.receiver.spells) do
		if(tostring(spell.spellid) == spellid and spell.class == classToken) then
			index = i
			break
		end
	end
	
	if(not index) then
		error("Unexpected error trying to delete spell from profile. Data consistency issue could mean more problems")
	end
	
	--delete from profile
	tremove(self.db.profile.receiver.spells, index)
	
	--delete from receiver
	--if(self:IsReceiving() == true) then
	local HermesDisplay = self:GetModule("HermesDisplay")
	HermesDisplay:RemoveSpell(spellid)
	--end
	
	--force the config window to update with latest spell info
	LIB_AceConfigRegistry:NotifyChange(AddonName)
	Hermes:BlizOptionsTable_Spells()
end



